 
// Select which PWM-capable pins are to be used.
int redPin = 9;
int greenPin = 3;
int bluePin = 10;


#include <SoftwareSerial.h>
//SoftwareSerial mySerial(rxPin, txPin)
//SoftwareSerial nss(2, 3);  //yellow wire = RX  Blue wire = TX
SoftwareSerial nss(0, 2);   //yellow wire = RX  Blue wire = TX


#include <avr/pgmspace.h>
#include <SSD1306.h>                  //for OLED
#include <SPI.h>                      //for OLED
#include <string.h>
#include <TinyGPS.h>

#define OLED_RESET 7
#define OLED_DC 8
#define OLED_CS 6     // SPI slave-select
#define OLED_MOSI 11   // hardware SPI MOSI
#define OLED_CLK 13    // hardware SPI clock





// Connect the GPS Power pin to 3.3V
// Connect the GPS Ground pin to ground
// Connect the GPS VBAT pin to 3.3V if no battery is used
// Connect the GPS TX (transmit) pin to Digital 2
// Connect the GPS RX (receive) pin to Digital 3
// For 3.3V only modules such as the UP501, connect a 10K
// resistor between digital 3 and GPS RX and a 10K resistor 
// from GPS RX to ground.
TinyGPS gps;
//NewSoftSerial mySerial(rxPin, txPin);

//OLED
SSD1306 oled(OLED_MOSI, OLED_CLK, OLED_DC, OLED_RESET, OLED_CS);


const int GPSActive = 4;               // Indicates GPS is getting data
const int statusLED = 5;               //System hearbeat LED

int updateFreq = 1000;                 //frequency to update GPS polls.  Good not to go below 1000 ms.

float flat, flon;                      //lat and long variables used in calculations
unsigned long age;                     //GPS reading age variable
float spd;                             // GPS Speed
float minDist = 100000;                //preset the minDist to some very high value (in meters)

//
//int md[] = {600, 180, 35};
//int j = 0;
//


const float ToRad = PI / 180.0;        // for HaverSine function 
const float R = 6371;                  // radius earth in Km - for HaverSine function
int mappedDist = 0;                    // mapped variable converted for placement of icons on OLED
int spdProRate = 0;                    // pro rated speed value movement of road graphic on OLED - to mimic speed changes
int i = 128;                           // used for the looping through animations of the OLED
int state = LOW;                       // The input state of the status LED
int gpsStatCounter = 0;
// setup variables for Tiny GPS
void gpsdump(TinyGPS &gps);
bool feedgps();

char cVal[32];
int sats;
int hdop;
String cardinal;


//********************************************************************************************************
// use for testing
boolean debug = true;            //print debugging statements to the serial port.
//********************************************************************************************************

// different commands to set the update rate from once a second (1 Hz) to 10 times a second (10Hz)
#define PMTK_SET_NMEA_UPDATE_1HZ  "$PMTK220,1000*1F"
#define PMTK_SET_NMEA_UPDATE_5HZ  "$PMTK220,200*2C"
#define PMTK_SET_NMEA_UPDATE_10HZ "$PMTK220,100*2F"
// turn on only the second sentence (GPRMC)
#define PMTK_SET_NMEA_OUTPUT_RMCONLY "$PMTK314,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*29"
// turn on ALL THE DATA
//#define PMTK_SET_NMEA_OUTPUT_ALLDATA "$PMTK314,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0*28"
// to generate your own sentences, check out the MTK command datasheet and use a checksum calculator
// such as the awesome http://www.hhhh.org/wiml/proj/nmeaxor.html



//Define list of waypoints to search
//format WayPt[Lat][Long]
//multidimensional array [howmany dn] [howmany across] 
const float wayPt[175][2] PROGMEM =
{
{43.25981,-79.88897},
{43.26085,-79.88347},
{43.2909,-79.8791},
{43.24039,-79.87665},
{43.2586,-79.84735},
{43.24956,-79.84735},
{43.2342,-79.84633},
{43.37923,-79.80605},
{43.19625,-79.79543},
{43.58813,-79.77104},
{43.69292,-79.7694},
{43.21895,-79.76782},
{43.61229,-79.75521},
{43.73482,-79.75486},
{43.70909,-79.73452},
{43.57842,-79.73367},
{43.63288,-79.72282},
{43.67563,-79.72268},
{43.48789,-79.71961},
{43.68061,-79.71723},
{43.75325,-79.70879},
{43.64605,-79.70614},
{43.55654,-79.70205},
{43.69511,-79.70133},
{43.7048,-79.69053},
{43.55075,-79.69002},
{43.73606,-79.68564},
{43.6656,-79.68487},
{43.54411,-79.67989},
{43.73315,-79.65815},
{43.70365,-79.64247},
{43.71626,-79.62724},
{43.74297,-79.60007},
{43.6867,-79.59548},
{43.6168,-79.59462},
{43.6093,-79.58389},
{43.74324,-79.58208},
{43.76426,-79.57475},
{43.69169,-79.57091},
{43.67451,-79.56318},
{43.61235,-79.56266},
{43.63564,-79.5623},
{43.6463,-79.56005},
{43.69429,-79.55863},
{43.61435,-79.55625},
{43.69711,-79.54668},
{43.68863,-79.54286},
{43.75209,-79.54239},
{43.59357,-79.53831},
{43.60718,-79.52076},
{43.75723,-79.51773},
{43.7001,-79.51627},
{43.74124,-79.51365},
{43.69014,-79.50059},
{43.74481,-79.48638},
{43.72655,-79.48203},
{43.70868,-79.47805},
{43.68207,-79.47158},
{43.78723,-79.47005},
{43.63524,-79.46814},
{43.6654,-79.46455},
{43.75029,-79.46335},
{43.75167,-79.45669},
{43.70696,-79.45317},
{43.73338,-79.45284},
{43.6564,-79.45248},
{43.69562,-79.45029},
{43.79209,-79.44757},
{43.71564,-79.44648},
{43.66669,-79.44646},
{43.67792,-79.44307},
{43.7738,-79.44298},
{43.75547,-79.43842},
{43.64058,-79.43666},
{43.65988,-79.43534},
{43.71927,-79.42978},
{43.79603,-79.42927},
{43.66198,-79.42567},
{43.79805,-79.42005},
{43.70234,-79.4191},
{43.67671,-79.41577},
{43.77975,-79.41558},
{43.66514,-79.41119},
{43.65646,-79.40769},
{43.7818,-79.4072},
{43.7441,-79.40673},
{43.7251,-79.40219},
{43.70673,-79.39839},
{43.64674,-79.39554},
{43.79228,-79.39358},
{43.70843,-79.39018},
{43.6575,-79.38963},
{43.76061,-79.38657},
{43.66489,-79.38456},
{43.64156,-79.38094},
{43.65166,-79.37898},
{43.64418,-79.3758},
{43.65703,-79.37455},
{43.65048,-79.37188},
{43.64819,-79.37092},
{43.79002,-79.36804},
{43.64903,-79.36726},
{43.65455,-79.36401},
{43.75292,-79.35998},
{43.6762,-79.35888},
{43.7346,-79.35596},
{43.79262,-79.35475},
{43.77511,-79.34706},
{43.74087,-79.34444},
{43.70748,-79.34361},
{43.72055,-79.33877},
{43.65513,-79.33757},
{43.68148,-79.33236},
{43.65922,-79.32844},
{43.81961,-79.32442},
{43.80636,-79.32098},
{43.82141,-79.31691},
{43.66551,-79.3162},
{43.72601,-79.31348},
{43.74212,-79.30982},
{43.71248,-79.30745},
{43.66872,-79.30578},
{43.77311,-79.30536},
{43.72472,-79.30266},
{43.79044,-79.30231},
{43.72571,-79.29753},
{43.70879,-79.29598},
{43.74539,-79.2947},
{43.8146,-79.29221},
{43.82828,-79.28717},
{43.72154,-79.28503},
{43.76583,-79.2811},
{43.72997,-79.27789},
{43.79837,-79.27339},
{43.77298,-79.27298},
{43.70701,-79.26813},
{43.69708,-79.26403},
{43.77011,-79.26128},
{43.75772,-79.23528},
{43.7764,-79.23172},
{43.75982,-79.22528},
{43.74324,-79.21892},
{43.78659,-79.1883},
{43.77062,-79.18696},
{43.7684,-79.18621},
{43.79898,-79.14617},
{45.32254,-75.88714},
{45.30397,-75.87684},
{45.36075,-75.79239},
{45.35086,-75.73362},
{45.36528,-75.73079},
{45.38925,-75.72662},
{45.41329,-75.71293},
{45.41903,-75.70239},
{45.40105,-75.69986},
{45.40881,-75.69373},
{45.37682,-75.69147},
{45.37674,-75.69135},
{45.4106,-75.68944},
{45.38881,-75.67722},
{45.35371,-75.64737},
{45.42424,-75.63732},
{45.41185,-75.63033},
{45.39286,-75.60757},
{45.46408,-75.54018},
{45.47731,-75.49738},
{45.49609,-73.65527},
{45.39268,-73.58067},
{45.49831,-73.56158},
{45.52936,-73.54607},
{46.0948,-71.30375},
{46.7847,-71.16127},
{43.37919101,-79.80605},
{43.449227,-79.704329},
{43.487886,-79.719564}
}; 


void setup()  
{
  
  Serial.begin(115200);
  pinMode(statusLED, OUTPUT); 
  
  //setup the RGB LED
  pinMode(redPin, OUTPUT);
  pinMode(greenPin, OUTPUT);
  pinMode(bluePin, OUTPUT);
  
  //set the RGB status LED to off which is HIGH (255) reason: Common-Annode.
  analogWrite(redPin, 255);
  analogWrite(greenPin, 255);
  analogWrite(bluePin, 255);

  
  //Serial.println("Adafruit MTK3329 NMEA test!");
  nss.begin(9600);  //needed to change this to 9600 baud for the UP501 GPS module
  
  //*********  OLED SETUP STUFF ******************************************************************/
  SPI.begin ();  //for OLED
  // OLED: By default, we'll generate the high voltage from the 3.3v line internally! (neat!)
  oled.ssd1306_init(SSD1306_SWITCHCAPVCC);

  //*********  GPS SETUP STUFF ******************************************************************/
  // 9600 NMEA is the default baud rate
  // uncomment this line to turn on only the "minimum recommended" data for high update rates!
  nss.println(PMTK_SET_NMEA_OUTPUT_RMCONLY);
  // uncomment this line to turn on all the available data - for 9600 baud you'll want 1 Hz rate
  //mySerial.println(PMTK_SET_NMEA_OUTPUT_ALLDATA);
  // Set the update rate
  // 1 Hz update rate
  nss.println(PMTK_SET_NMEA_UPDATE_1HZ);
  // 5 Hz update rate- for 9600 baud you'll have to set the output to RMC only (see above)
  //nss.println(PMTK_SET_NMEA_UPDATE_5HZ);
  // 10 Hz update rate - for 9600 baud you'll have to set the output to RMC only (see above)
  //nss.println(PMTK_SET_NMEA_UPDATE_10HZ);
  //*********  END GPS SETUP STUFF ******************************************************************/

  //setupDisplay();
  
  
}


void loop()
{
  bool newdata = false;
  unsigned long start = millis();


  blinkLED();   // heartbeat to ensure system is working
  flat=0;
  flon=0;
  spd=0;
  sats=0;
  hdop=0;
  cardinal= "";
  
  
  // Every xxxx milliseconds we print an update
  while (millis() - start < updateFreq)
  {
    if (feedgps())
      newdata = true;    
  }
  
  heartbeat(newdata);  //shows external heartbeat signal through GPS status LED
       
      if (newdata)
      {
                  feedgps(); // If we don't feed the gps during this long routine, we may drop characters and get checksum errors
                  gps.f_get_position(&flat, &flon, &age);
                  if (debug) {
                    Serial.print("Lat/Long(float): "); Serial.print(flat, 5); Serial.print(", "); Serial.print(flon, 5);
                    Serial.print(" Fix age: "); Serial.print(age); Serial.print("ms.");
                    Serial.print(" (kmph): "); Serial.print(gps.f_speed_kmph(),4); Serial.println();
                    //Serial.print(" Sats: "); Serial.print(gps.satellites()); Serial.println();
                    feedgps(); // If we don't feed the gps during this long routine, we may drop characters and get checksum errors
                    Serial.print(" Sats: "); print_int(gps.satellites(), TinyGPS::GPS_INVALID_SATELLITES, 5);
                    Serial.print(" HDOP: "); print_int(gps.hdop(), TinyGPS::GPS_INVALID_HDOP, 5);
                    Serial.print(" Course: "); print_float(gps.f_course(), TinyGPS::GPS_INVALID_F_ANGLE, 7, 2);
                    Serial.print(" Cardinal: "); print_str(gps.f_course() == TinyGPS::GPS_INVALID_F_ANGLE ? "*** " : TinyGPS::cardinal(gps.f_course()), 6);

                  }
                  //compare the latest reading with the array of red light camera points and find which is closest.
                  minDist = getClosestPt(flat, flon);
                  //minDist = md[j];
                  //if (j == 2){j = 0;}else{j++;}
                
                  

                  
                  //indicate that GPS is active and receiving data
                  flash(1);  //flashes the LED on the project box showing GPS signal
                 
          
          
                  if (debug) {
                    Serial.print("Delta = ");Serial.print(minDist, 2);Serial.println(" Meters   ");
                  }

                  //Process alerts via various means based on pre-defined distance thresholds.
                  
                  if (minDist < 50)                 //Warning Alert < 50 meters - in range if red light camera
                  {
                    //red
                    changeWarning(4);
                  }
                  else if (minDist < 200)          //Approaching Alert < 200 meters
                  {
                    //yellow
                    changeWarning(3);
                  }
                  else if (minDist <= 600)          //Initial Alert - 600 meters
                  {
                    //blue - early warning
                    changeWarning(2);
                  }
                  else 
                  {
                    //just show status that GPS is working.
                    changeWarning(1);
                  }
                  
                  //for testing
                  //minDist = 40;
                  //updateDisplay(minDist, 65.00);
                  //end testing
                  
                  //updateDisplay(minDist, gps.f_speed_kmph());
          
      }
      
      else
      {
          if (debug) {Serial.println("Polling GPS... waiting for fix...");}
      }
 
 


      oled.clear();

      //Distance to next red light
      oled.drawstring(0, 0, "Dist:");
      if (minDist == 100000)
      {
        oled.drawstring(30, 0, "?");
      }
        else
      {
        ftoa(cVal, minDist/1000, 2);
        oled.drawstring(30, 0, cVal);
      }
      oled.drawstring(55, 0, "km");


      //Cardinal Direction of vehicle
      //strcpy ( cVal, TinyGPS::cardinal(gps.f_course()) );
      strcpy ( cVal, gps.f_course() == TinyGPS::GPS_INVALID_F_ANGLE ? "?" : TinyGPS::cardinal(gps.f_course()));
      oled.drawstring(75, 0, "Dir:");
      oled.drawstring(100, 0, cVal);
      
      oled.drawline(0, 12, 128, 12, WHITE);

      //speed
      spd = gps.f_speed_kmph();
      if (spd < 0) {spd=0;}
      ftoa(cVal, spd, 0);
      oled.drawstring(0, 2, "Spd:");
      oled.drawstring(25, 2, cVal);


      
      //satellites
      itoa(gps.satellites(), cVal, 0);
      oled.drawstring(65, 2, "Sats:");
      oled.drawstring(90, 2, cVal);

      

      //latitude      
      ftoa(cVal, flat, 2);
      oled.drawstring(0, 3, "Lat:");
      oled.drawstring(25, 3, cVal);
      //longitude
      ftoa(cVal, flon, 2);
      oled.drawstring(65, 3, "Lng:");
      oled.drawstring(90, 3, cVal);


      //Display all the results
      oled.display(); 

      //reset minDist for the next round
      minDist = 100000;     //meters  

       
}






//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
char *ftoa(char *a, double f, int precision)
{
  long p[] = {0,10,100,1000,10000,100000,1000000,10000000,100000000};
  
  char *ret = a;
  long heiltal = (long)f;
  itoa(heiltal, a, 10);
  while (*a != '\0') a++;
  *a++ = '.';
  long desimal = abs((long)((f - heiltal) * p[precision]));
  itoa(desimal, a, 10);
  return ret;
}
 
 
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void setupDisplay()
{
    // Draw animated graphics to OLED - we will use the GPS speed to mimic changes in speed of the lanes.
    oled.clear();   // clears the screen and buffer 
    //oled.drawbitmap(0, 2,  Road, 128, 32, 1);     //road - dash lines - draw the animating road
    //oled.drawbitmap(95, 4,  JukeIcon, 34, 40, 1);  //keep drawing the car icon in place
    oled.fillrect(0, 0, 128, 2, WHITE);            //keep drawing the top continuous line
    oled.fillrect(0, 30, 128, 2, WHITE);           //keep drawing the bottom continous line
    oled.display(); 
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void updateDisplay(int minDist, float spd)
{
    if (i==0){i=128;}  //we have 128 pixels to loop through for animation of the road dashed lines. reset when it gets to 0.
    
    // Draw animated graphics to OLED - we will use the GPS speed to mimic changes in speed of the lanes.
    //oled.drawbitmap(i, 16,  Road, 128, 32, 1);     //road - dash lines - draw the animating road
    //oled.drawbitmap(i-128, 16,  Road, 128, 32, 1);  //road - dash lines - draw the road ahead to keep contents on the screen
 
    //map values based on dist from camera, min dist on ground of 20m, max dist on ground of 500m, min dist in px to car, max dist in px from car
    mappedDist = map(minDist, 20, 500, 40, 110);
    //oled.drawbitmap(mappedDist, 10,  RedLightCam, 10, 15, 1);  //move the red light camera as it approaches 
  
    //oled.drawbitmap(95, 12,  JukeIcon, 34, 40, 1);  //keep drawing the car icon in place
    oled.fillrect(0, 0, 128, 4, WHITE);            //keep drawing the top continuous line
    oled.fillrect(0, 61, 128, 4, WHITE);           //keep drawing the bottom continous line
    oled.display();
    oled.clear();   // clears the screen and buffer 
  

    // convert speed to an int so it can be used in the map funciton.  1.25 km/h = 125
    //Serial.println(int(spd*10));
    spdProRate = map(int(spd*10), 0, 900, 0, 30);
    //Serial.println(spdProRate);
    //i=i-10; 
    i = i - spdProRate;
    if (i < 0) {i=0;}
    
    //mappedDist = 0;
}



//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
float getClosestPt(float flat, float flon)
{
  float d;
  float d2;
  float md = 10000;
  //int minCoordIndex;
  //int coordsProcessed;

              if (flat != 0 && flon != 0)
              {
                for (unsigned int x=0; x<175; x++){
                      //In array lat is 2nd col, long is 1st col  e.g. lat:43.xxxx,long:-79.xxxxx - need to get right ones for calc
                      //flat2 =  pgm_read_float_near(&wayPt[x][0]);
                      //flon2 = pgm_read_float_near(&wayPt[x][1]);
                      //if (debug)
                      //{
                      //  Serial.print("Flat2:  ");Serial.print(flat2, 5);
                      //  Serial.print("  Flon2:  ");Serial.println(flon2, 5);
                      //}
                      //d = HaverSine(flat, flon, pgm_read_float_near(&wayPt[x][0]), pgm_read_float_near(&wayPt[x][1])); 
                      d = gps.distance_between(flat, flon, pgm_read_float_near(&wayPt[x][0]), pgm_read_float_near(&wayPt[x][1]))/1000 ;
                      if (debug)
                      {
                        //Serial.print("DistanceCalcualted d1: ");Serial.print(d, 4); Serial.print(" d2:"); Serial.print(d2, 4); Serial.println("");  
                        //Serial.print("DistanceCalcualted d1: ");Serial.print(d, 4); Serial.println("");  
                      }
                      if (d < md)
                      {
                        md = d;
                        //minCoordIndex = x;
                      }
                      //coordsProcessed++;            
                  }
          
                  //convert to meters
                  md = md * 1000;
                  
                  //Serial.print(" Processed:");Serial.println(coordsProcessed);
                  /*
                  Serial.println("");
                  Serial.println("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
                  Serial.print(" Index:");Serial.println(minCoordIndex);
                  Serial.print(" Processed:");Serial.println(coordsProcessed);
                  Serial.println("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
                  Serial.println("");
                  */
                }
                //coordsProcessed = 0;
 return md; 
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 float HaverSine(float lat1, float lon1, float lat2, float lon2)
{
  //float ToRad = PI / 180.0;
  //float R = 6371;   // radius earth in Km
  
  float dLat = (lat2-lat1) * ToRad;
  float dLon = (lon2-lon1) * ToRad;

  float a = sin(dLat/2) * sin(dLat/2) +
	  cos(lat1 * ToRad) * cos(lat2 * ToRad) *
	  sin(dLon/2) * sin(dLon/2);


  float c = 2 * atan2(sqrt(a), sqrt(1-a));

  float d = R * c;
  return d;  //distance returned in Kilometers
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  
bool feedgps()
{
  while (nss.available())
  {
    if (gps.encode(nss.read()))
      return true;
  }
  return false;
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  
static void print_float(float val, float invalid, int len, int prec)
{
  char sz[32];
  if (val == invalid)
  {
    strcpy(sz, "*******");
    sz[len] = 0;
        if (len > 0) 
          sz[len-1] = ' ';
    for (int i=7; i<len; ++i)
        sz[i] = ' ';
    Serial.print(sz);
  }
  else
  {
    Serial.print(val, prec);
    int vi = abs((int)val);
    int flen = prec + (val < 0.0 ? 2 : 1);
    flen += vi >= 1000 ? 4 : vi >= 100 ? 3 : vi >= 10 ? 2 : 1;
    for (int i=flen; i<len; ++i)
      Serial.print(" ");
  }
  feedgps();
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
/*
void beep(int howmany, int itone) {
  
 if (itone == 0){itone=2200;}
  
 for (int i=0; i < howmany; i++){
    tone(piezo, itone, 10);
    delay(100);
  }
}
*/
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
static void print_str(const char *str, int len)
{
  int slen = strlen(str);
  for (int i=0; i<len; ++i)
    Serial.print(i<slen ? str[i] : ' ');
  feedgps();
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void flash(int howmany) {
 for (int i=0; i < howmany; i++){
    digitalWrite(GPSActive, HIGH);
    delay(100);
    digitalWrite(GPSActive, LOW);
    delay(100);
  }
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void heartbeat(boolean nd){
    
  //show if there is no data - hearbeat of system - shows it is working.  Fast blink.
  if (nd == false){

    if (gpsStatCounter == 2)
      {
          analogWrite(redPin, 127);   
          analogWrite(greenPin, 255);   
          analogWrite(bluePin, 0);   
          delay(50); 
          analogWrite(redPin, 255);             
          analogWrite(greenPin, 255);             
          analogWrite(bluePin, 255);             
          //delay(50);     
          
          gpsStatCounter = 0;
      }
      else
      {
      gpsStatCounter++;
      }
  }  

}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void changeWarning(int stateLevel)
{
    switch (stateLevel) {
      
    case 1:
      //green LED - show for status that GPS is getting data
      analogWrite(greenPin, 155);
      delay(200);
      break;

  
     case 2:
      //early warning.
      analogWrite(bluePin, 155);
      delay(200);
      break;
    
    case 3:
      //yellow
      analogWrite(redPin, 0);
      analogWrite(greenPin, 190);
      delay(200);      
      break;
      
    case 4:
      //in and just before red light camera
      analogWrite(redPin, 155);  
      delay(200); 
      break;  
        
    default:
      //all off
      analogWrite(greenPin, 255);
      analogWrite(bluePin, 255);
      analogWrite(redPin, 255);    
      break;    
    }
    
    

    //turn the LEDs back off
    analogWrite(greenPin, 255);
    analogWrite(bluePin, 255);
    analogWrite(redPin, 255);    


  
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
static void print_int(unsigned long val, unsigned long invalid, int len)
{
  char sz[32];
  if (val == invalid)
    strcpy(sz, "*******");
  else
    sprintf(sz, "%ld", val);
  sz[len] = 0;
  for (int i=strlen(sz); i<len; ++i)
    sz[i] = ' ';
  if (len > 0) 
    sz[len-1] = ' ';
  Serial.print(sz);
  feedgps();
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void blinkLED()
{
state = !state;
digitalWrite(statusLED, state); 
}


